# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import logging
from daCore import BasicObjects, PlatformInfo
import numpy, copy
mpr = PlatformInfo.PlatformInfo().MachinePrecision()
mfp = PlatformInfo.PlatformInfo().MaximumPrecision()

# ==============================================================================
class ElementaryAlgorithm(BasicObjects.Algorithm):
    def __init__(self):
        BasicObjects.Algorithm.__init__(self, "PARALLELFUNCTIONTEST")
        self.defineRequiredParameter(
            name     = "NumberOfPrintedDigits",
            default  = 5,
            typecast = int,
            message  = "Nombre de chiffres affichés pour les impressions de réels",
            minval   = 0,
            )
        self.defineRequiredParameter(
            name     = "NumberOfRepetition",
            default  = 1,
            typecast = int,
            message  = "Nombre de fois où l'exécution de la fonction est répétée",
            minval   = 1,
            )
        self.defineRequiredParameter(
            name     = "ResultTitle",
            default  = "",
            typecast = str,
            message  = "Titre du tableau et de la figure",
            )
        self.defineRequiredParameter(
            name     = "SetDebug",
            default  = False,
            typecast = bool,
            message  = "Activation du mode debug lors de l'exécution",
            )
        self.defineRequiredParameter(
            name     = "StoreSupplementaryCalculations",
            default  = [],
            typecast = tuple,
            message  = "Liste de calculs supplémentaires à stocker et/ou effectuer",
            listval  = [
                "CurrentState",
                "SimulatedObservationAtCurrentState",
                ]
            )
        self.requireInputArguments(
            mandatory= ("Xb", "HO"),
            )
        self.setAttributes(tags=(
            "Checking",
            ))

    def run(self, Xb=None, Y=None, U=None, HO=None, EM=None, CM=None, R=None, B=None, Q=None, Parameters=None):
        self._pre_run(Parameters, Xb, Y, U, HO, EM, CM, R, B, Q)
        #
        Hm = HO["Direct"].appliedTo
        #
        Xn = copy.copy( Xb )
        #
        # ----------
        __marge =  5*u" "
        _p = self._parameters["NumberOfPrintedDigits"]
        if len(self._parameters["ResultTitle"]) > 0:
            __rt = str(self._parameters["ResultTitle"])
            msgs  = u"\n"
            msgs +=  __marge + "====" + "="*len(__rt) + "====\n"
            msgs +=  __marge + "    " + __rt + "\n"
            msgs +=  __marge + "====" + "="*len(__rt) + "====\n"
            print("%s"%msgs)
        #
        msgs  = ("===> Information before launching:\n")
        msgs += ("     -----------------------------\n")
        msgs += ("     Characteristics of input vector X, internally converted:\n")
        msgs += ("       Type...............: %s\n")%type( Xn )
        msgs += ("       Lenght of vector...: %i\n")%max(numpy.asarray( Xn ).shape)
        msgs += ("       Minimum value......: %."+str(_p)+"e\n")%numpy.min( Xn )
        msgs += ("       Maximum value......: %."+str(_p)+"e\n")%numpy.max( Xn )
        msgs += ("       Mean of vector.....: %."+str(_p)+"e\n")%numpy.mean( Xn, dtype=mfp )
        msgs += ("       Standard error.....: %."+str(_p)+"e\n")%numpy.std( Xn, dtype=mfp )
        msgs += ("       L2 norm of vector..: %."+str(_p)+"e\n")%numpy.linalg.norm( Xn )
        print(msgs)
        #
        print("     %s\n"%("-"*75,))
        if self._parameters["SetDebug"]:
            CUR_LEVEL = logging.getLogger().getEffectiveLevel()
            logging.getLogger().setLevel(logging.DEBUG)
            print("===> Beginning of evaluation, activating debug\n")
        else:
            print("===> Beginning of evaluation, without activating debug\n")
        #
        Xs = []
        Ys = []
        for i in range(self._parameters["NumberOfRepetition"]):
            if self._toStore("CurrentState"):
                self.StoredVariables["CurrentState"].store( numpy.ravel(Xn) )
            Xs.append( Xn )
        #
        # ----------
        HO["Direct"].disableAvoidingRedundancy()
        # ----------
        Ys = Hm( Xs, argsAsSerie = True )
        # ----------
        HO["Direct"].enableAvoidingRedundancy()
        # ----------
        #
        print()
        if self._parameters["SetDebug"]:
            print("===> End of evaluation, deactivating debug\n")
            logging.getLogger().setLevel(CUR_LEVEL)
        else:
            print("===> End of evaluation, without deactivating debug\n")
        #
        for i in range(self._parameters["NumberOfRepetition"]):
            print("     %s\n"%("-"*75,))
            if self._parameters["NumberOfRepetition"] > 1:
                print("===> Repetition step number %i on a total of %i\n"%(i+1,self._parameters["NumberOfRepetition"]))
            #
            Yn = Ys[i]
            msgs  = ("===> Information after evaluation:\n")
            msgs += ("\n     Characteristics of simulated output vector Y=H(X), to compare to others:\n")
            msgs += ("       Type...............: %s\n")%type( Yn )
            msgs += ("       Lenght of vector...: %i\n")%max(numpy.asarray( Yn ).shape)
            msgs += ("       Minimum value......: %."+str(_p)+"e\n")%numpy.min( Yn )
            msgs += ("       Maximum value......: %."+str(_p)+"e\n")%numpy.max( Yn )
            msgs += ("       Mean of vector.....: %."+str(_p)+"e\n")%numpy.mean( Yn, dtype=mfp )
            msgs += ("       Standard error.....: %."+str(_p)+"e\n")%numpy.std( Yn, dtype=mfp )
            msgs += ("       L2 norm of vector..: %."+str(_p)+"e\n")%numpy.linalg.norm( Yn )
            print(msgs)
            if self._toStore("SimulatedObservationAtCurrentState"):
                self.StoredVariables["SimulatedObservationAtCurrentState"].store( numpy.ravel(Yn) )
        #
        if self._parameters["NumberOfRepetition"] > 1:
            msgs  = ("     %s\n"%("-"*75,))
            msgs += ("\n===> Statistical analysis of the outputs obtained through parallel repeated evaluations\n")
            msgs += ("\n     (Remark: numbers that are (about) under %.0e represent 0 to machine precision)\n"%mpr)
            Yy = numpy.array( Ys )
            msgs += ("\n     Characteristics of the whole set of outputs Y:\n")
            msgs += ("       Number of evaluations.........................: %i\n")%len( Ys )
            msgs += ("       Minimum value of the whole set of outputs.....: %."+str(_p)+"e\n")%numpy.min( Yy )
            msgs += ("       Maximum value of the whole set of outputs.....: %."+str(_p)+"e\n")%numpy.max( Yy )
            msgs += ("       Mean of vector of the whole set of outputs....: %."+str(_p)+"e\n")%numpy.mean( Yy, dtype=mfp )
            msgs += ("       Standard error of the whole set of outputs....: %."+str(_p)+"e\n")%numpy.std( Yy, dtype=mfp )
            Ym = numpy.mean( numpy.array( Ys ), axis=0, dtype=mfp )
            msgs += ("\n     Characteristics of the vector Ym, mean of the outputs Y:\n")
            msgs += ("       Size of the mean of the outputs...............: %i\n")%Ym.size
            msgs += ("       Minimum value of the mean of the outputs......: %."+str(_p)+"e\n")%numpy.min( Ym )
            msgs += ("       Maximum value of the mean of the outputs......: %."+str(_p)+"e\n")%numpy.max( Ym )
            msgs += ("       Mean of the mean of the outputs...............: %."+str(_p)+"e\n")%numpy.mean( Ym, dtype=mfp )
            msgs += ("       Standard error of the mean of the outputs.....: %."+str(_p)+"e\n")%numpy.std( Ym, dtype=mfp )
            Ye = numpy.mean( numpy.array( Ys ) - Ym, axis=0, dtype=mfp )
            msgs += "\n     Characteristics of the mean of the differences between the outputs Y and their mean Ym:\n"
            msgs += ("       Size of the mean of the differences...........: %i\n")%Ym.size
            msgs += ("       Minimum value of the mean of the differences..: %."+str(_p)+"e\n")%numpy.min( Ye )
            msgs += ("       Maximum value of the mean of the differences..: %."+str(_p)+"e\n")%numpy.max( Ye )
            msgs += ("       Mean of the mean of the differences...........: %."+str(_p)+"e\n")%numpy.mean( Ye, dtype=mfp )
            msgs += ("       Standard error of the mean of the differences.: %."+str(_p)+"e\n")%numpy.std( Ye, dtype=mfp )
            msgs += ("\n     %s\n"%("-"*75,))
            print(msgs)
        #
        self._post_run(HO)
        return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



#-*-coding:utf-8-*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import SALOMERuntime
import pilot
try:
    import cPickle as pickle
except:
    import pickle
import numpy
import threading
import sys
import traceback
import codecs

# Pour disposer des classes dans l'espace de nommage lors du pickle
from daCore.Aidsm import Aidsm
from daCore.AssimilationStudy import AssimilationStudy # JPA à résorber
from daYacsIntegration import daStudy

def dumps( data ):
  return str(codecs.encode(pickle.dumps(data), "base64").decode())
def loads( data ):
  return pickle.loads(codecs.decode(data.encode(), "base64"))

class OptimizerHooks:

  def __init__(self, optim_algo, switch_value=-1):
    self.optim_algo = optim_algo
    self.switch_value = str(int(switch_value))

  def create_sample(self, data, method):
    sample = pilot.StructAny_New(self.optim_algo.runtime.getTypeCode('SALOME_TYPES/ParametricInput'))

    # TODO Input, Output VarList
    inputVarList  = pilot.SequenceAny_New(self.optim_algo.runtime.getTypeCode("string"))
    outputVarList = pilot.SequenceAny_New(self.optim_algo.runtime.getTypeCode("string"))
    for var in self.optim_algo.da_study.InputVariables:
      inputVarList.pushBack(var)
    for var in self.optim_algo.da_study.OutputVariables:
      outputVarList.pushBack(var)
    sample.setEltAtRank("inputVarList", inputVarList)
    sample.setEltAtRank("outputVarList", outputVarList)

    # Les parametres specifiques à ADAO
    specificParameters = pilot.SequenceAny_New(self.optim_algo.runtime.getTypeCode("SALOME_TYPES/Parameter"))
    method_name = pilot.StructAny_New(self.optim_algo.runtime.getTypeCode('SALOME_TYPES/Parameter'))
    method_name.setEltAtRank("name", "method")
    method_name.setEltAtRank("value", dumps(method))
    specificParameters.pushBack(method_name)
    # print self.optim_algo.has_observer
    if self.optim_algo.has_observer:
      obs_switch = pilot.StructAny_New(self.optim_algo.runtime.getTypeCode('SALOME_TYPES/Parameter'))
      obs_switch.setEltAtRank("name", "switch_value")
      obs_switch.setEltAtRank("value", dumps("1"))
      specificParameters.pushBack(obs_switch)
    if self.optim_algo.has_evolution_model:
      obs_switch = pilot.StructAny_New(self.optim_algo.runtime.getTypeCode('SALOME_TYPES/Parameter'))
      obs_switch.setEltAtRank("name", "switch_value")
      obs_switch.setEltAtRank("value", dumps(self.switch_value))
      specificParameters.pushBack(obs_switch)
    sample.setEltAtRank("specificParameters", specificParameters)

    # Les données
    # TODO à faire
    #print data
    #print data.ndim
    #print data.shape
    #print data[:,0]
    #print data.flatten()
    #print data.flatten().shape

    variable          = pilot.SequenceAny_New(self.optim_algo.runtime.getTypeCode("double"))
    variable_sequence = pilot.SequenceAny_New(variable.getType())
    state_sequence    = pilot.SequenceAny_New(variable_sequence.getType())
    time_sequence     = pilot.SequenceAny_New(state_sequence.getType())

    #print "Input Data", data
    if isinstance(data, type((1,2))):
      self.add_parameters(data[0], variable_sequence)
      self.add_parameters(data[1], variable_sequence, Output=True) # Output == Y
    else:
      self.add_parameters(data, variable_sequence)
    state_sequence.pushBack(variable_sequence)
    time_sequence.pushBack(state_sequence)
    sample.setEltAtRank("inputValues", time_sequence)
    return sample

  def add_parameters(self, data, variable_sequence, Output=False):
    param = pilot.SequenceAny_New(self.optim_algo.runtime.getTypeCode("double"))
    elt_list = 0 # index dans la liste des arguments
    val_number = 0 # nbre dans l'argument courant
    if not Output:
      val_end = self.optim_algo.da_study.InputVariables[self.optim_algo.da_study.InputVariablesOrder[elt_list]] # nbr de l'argument courant (-1 == tout)
    else:
      val_end = self.optim_algo.da_study.OutputVariables[self.optim_algo.da_study.OutputVariablesOrder[elt_list]] # nbr de l'argument courant (-1 == tout)

    if data is None:
        it = []
    else:
        it = data.flat
    for val in it:
      param.pushBack(val)
      val_number += 1
      # Test si l'argument est ok
      if val_end != -1:
        if val_number == val_end:
          variable_sequence.pushBack(param)
          param = pilot.SequenceAny_New(self.optim_algo.runtime.getTypeCode("double"))
          val_number = 0
          elt_list += 1
          if not Output:
            if elt_list < len(self.optim_algo.da_study.InputVariablesOrder):
              val_end = self.optim_algo.da_study.InputVariables[self.optim_algo.da_study.InputVariablesOrder[elt_list]]
            else:
              break
          else:
            if elt_list < len(self.optim_algo.da_study.OutputVariablesOrder):
              val_end = self.optim_algo.da_study.OutputVariables[self.optim_algo.da_study.OutputVariablesOrder[elt_list]]
            else:
              break
    if val_end == -1:
      variable_sequence.pushBack(param)

  def get_data_from_any(self, any_data):
    error = any_data["returnCode"].getIntValue()
    if error != 0:
      self.optim_algo.setError(any_data["errorMessage"].getStringValue())

    data = []
    outputValues = any_data["outputValues"]
    #print outputValues
    for variable in outputValues[0][0]:
      for i in range(variable.size()):
        data.append(variable[i].getDoubleValue())

    matrix = numpy.matrix(data).T
    return matrix

  def Direct(self, X, sync = 1):
    # print "Call Direct OptimizerHooks"
    if sync == 1:
      # 1: Get a unique sample number
      self.optim_algo.counter_lock.acquire()
      self.optim_algo.sample_counter += 1
      local_counter = self.optim_algo.sample_counter

      # 2: Put sample in the job pool
      sample = self.create_sample(X, "Direct")
      self.optim_algo.pool.pushInSample(local_counter, sample)

      # 3: Wait
      while True:
        #print "waiting"
        self.optim_algo.signalMasterAndWait()
        #print "signal"
        if self.optim_algo.isTerminationRequested():
          self.optim_algo.pool.destroyAll()
          return
        else:
          # Get current Id
          sample_id = self.optim_algo.pool.getCurrentId()
          if sample_id == local_counter:
            # 4: Data is ready
            any_data = self.optim_algo.pool.getOutSample(local_counter)
            Z = self.get_data_from_any(any_data)

            # 5: Release lock
            # Have to be done before but need a new implementation
            # of the optimizer loop
            self.optim_algo.counter_lock.release()
            return Z
    else:
      #print "sync false is not yet implemented"
      self.optim_algo.setError("sync == false not yet implemented")

  def Tangent(self, xxx_todo_changeme, sync = 1):
    # print "Call Tangent OptimizerHooks"
    (X, dX) = xxx_todo_changeme
    if sync == 1:
      # 1: Get a unique sample number
      self.optim_algo.counter_lock.acquire()
      self.optim_algo.sample_counter += 1
      local_counter = self.optim_algo.sample_counter

      # 2: Put sample in the job pool
      sample = self.create_sample((X,dX) , "Tangent")
      self.optim_algo.pool.pushInSample(local_counter, sample)

      # 3: Wait
      while True:
        self.optim_algo.signalMasterAndWait()
        if self.optim_algo.isTerminationRequested():
          self.optim_algo.pool.destroyAll()
          return
        else:
          # Get current Id
          sample_id = self.optim_algo.pool.getCurrentId()
          if sample_id == local_counter:
            # 4: Data is ready
            any_data = self.optim_algo.pool.getOutSample(local_counter)
            Z = self.get_data_from_any(any_data)

            # 5: Release lock
            # Have to be done before but need a new implementation
            # of the optimizer loop
            self.optim_algo.counter_lock.release()
            return Z
    else:
      #print "sync false is not yet implemented"
      self.optim_algo.setError("sync == false not yet implemented")

  def Adjoint(self, xxx_todo_changeme1, sync = 1):
    # print "Call Adjoint OptimizerHooks"
    (X, Y) = xxx_todo_changeme1
    if sync == 1:
      # 1: Get a unique sample number
      self.optim_algo.counter_lock.acquire()
      self.optim_algo.sample_counter += 1
      local_counter = self.optim_algo.sample_counter

      # 2: Put sample in the job pool
      sample = self.create_sample((X,Y), "Adjoint")
      self.optim_algo.pool.pushInSample(local_counter, sample)

      # 3: Wait
      while True:
        #print "waiting"
        self.optim_algo.signalMasterAndWait()
        #print "signal"
        if self.optim_algo.isTerminationRequested():
          self.optim_algo.pool.destroyAll()
          return
        else:
          # Get current Id
          sample_id = self.optim_algo.pool.getCurrentId()
          if sample_id == local_counter:
            # 4: Data is ready
            any_data = self.optim_algo.pool.getOutSample(local_counter)
            Z = self.get_data_from_any(any_data)

            # 5: Release lock
            # Have to be done before but need a new implementation
            # of the optimizer loop
            self.optim_algo.counter_lock.release()
            return Z
    else:
      #print "sync false is not yet implemented"
      self.optim_algo.setError("sync == false not yet implemented")

class AssimilationAlgorithm_asynch(SALOMERuntime.OptimizerAlgASync):

  def __init__(self):
    SALOMERuntime.RuntimeSALOME_setRuntime()
    SALOMERuntime.OptimizerAlgASync.__init__(self, None)
    self.runtime = SALOMERuntime.getSALOMERuntime()

    self.has_evolution_model = False
    self.has_observer = False

    # Gestion du compteur
    self.sample_counter = 0
    self.counter_lock = threading.Lock()

    # Definission des types d'entres et de sorties pour le code de calcul
    self.tin      = self.runtime.getTypeCode("SALOME_TYPES/ParametricInput")
    self.tout     = self.runtime.getTypeCode("SALOME_TYPES/ParametricOutput")
    self.pyobject = self.runtime.getTypeCode("pyobj")

    # Absolument indispensable de definir ainsi "self.optim_hooks"
    # (sinon on a une "Unknown Exception" sur l'attribut "finish")
    self.optim_hooks = OptimizerHooks(self)

  # input vient du port algoinit, input est un Any YACS !
  def initialize(self,input):
    #print "Algorithme initialize"

    # get the daStudy
    # print "[Debug] Input is ", input
    if sys.version_info.major < 3:
        str_da_study = input.getStringValue()
    else:
        str_da_study = input.getBytes()
    try:
        self.da_study = pickle.loads(str_da_study)
    except ValueError as e:
        raise ValueError("\n\n  Handling internal error in study exchange (message: \"%s\").\n  The case is probably too big (bigger than the physical plus the virtual memory available).\n  Try if possible to store the covariance matrices in sparse format.\n"%(str(e),))
    #print "[Debug] da_study is ", self.da_study
    self.da_study.initYIAlgorithm()
    self.ADD = self.da_study.getResults()

  def startToTakeDecision(self):
    # print "Algorithme startToTakeDecision"

    # Check if ObservationOperator is already set
    if self.da_study.getYIObservationOperatorType("Direct") == "Function" or self.da_study.getYIObservationOperatorType("Tangent") == "Function" or self.da_study.getYIObservationOperatorType("Adjoint") == "Function" :
      # print "Set Hooks for ObservationOperator"
      # Use proxy function for YACS
      self.hooksOO = OptimizerHooks(self, switch_value=1)
      direct = tangent = adjoint = None
      if self.da_study.getYIObservationOperatorType("Direct") == "Function":
        direct = self.hooksOO.Direct
      if self.da_study.getYIObservationOperatorType("Tangent") == "Function" :
        tangent = self.hooksOO.Tangent
      if self.da_study.getYIObservationOperatorType("Adjoint") == "Function" :
        adjoint = self.hooksOO.Adjoint

      # Set ObservationOperator
      self.ADD.setObservationOperator(ThreeFunctions = {"Direct":direct, "Tangent":tangent, "Adjoint":adjoint})
    # else:
      # print "Not setting Hooks for ObservationOperator"

    # Check if EvolutionModel is already set
    if self.da_study.getYIEvolutionModelType("Direct") == "Function" or self.da_study.getYIEvolutionModelType("Tangent") == "Function" or self.da_study.getYIEvolutionModelType("Adjoint") == "Function" :
      self.has_evolution_model = True
      # print "Set Hooks for EvolutionModel"
      # Use proxy function for YACS
      self.hooksEM = OptimizerHooks(self, switch_value=2)
      direct = tangent = adjoint = None
      if self.da_study.getYIEvolutionModelType("Direct") == "Function":
        direct = self.hooksEM.Direct
      if self.da_study.getYIEvolutionModelType("Tangent") == "Function" :
        tangent = self.hooksEM.Tangent
      if self.da_study.getYIEvolutionModelType("Adjoint") == "Function" :
        adjoint = self.hooksEM.Adjoint

      # Set EvolutionModel
      self.ADD.setEvolutionModel(ThreeFunctions = {"Direct":direct, "Tangent":tangent, "Adjoint":adjoint})
    # else:
      # print "Not setting Hooks for EvolutionModel"

    # Set Observers
    for observer_name in list(self.da_study.observers_dict.keys()):
      # print "observers %s found" % observer_name
      self.has_observer = True
      if self.da_study.observers_dict[observer_name]["scheduler"] != "":
        self.ADD.setObserver(Variable = observer_name, ObjectFunction = self.obs, Scheduler = self.da_study.observers_dict[observer_name]["scheduler"], Info = observer_name)
      else:
        self.ADD.setObserver(Variable = observer_name, ObjectFunction = self.obs, Info = observer_name)

    # Start Assimilation Study
    print("Launching the analysis\n")
    try:
        self.ADD.execute()
    except Exception as e:
        if isinstance(e, type(SyntaxError())): msg = "at %s: %s"%(e.offset, e.text)
        else: msg = ""
        raise ValueError("during execution, the following error occurs:\n\n%s %s\n\nSee also the potential messages, which can show the origin of the above error, in the YACS GUI or in the launching terminal."%(str(e),msg))

    # Assimilation Study is finished
    self.pool.destroyAll()

  def obs(self, var, info):
    # print("Call observer %s with var type %s" %(info,type(var))
    sample = pilot.StructAny_New(self.runtime.getTypeCode('SALOME_TYPES/ParametricInput'))

    # Fake data
    inputVarList  = pilot.SequenceAny_New(self.runtime.getTypeCode("string"))
    outputVarList = pilot.SequenceAny_New(self.runtime.getTypeCode("string"))
    inputVarList.pushBack("a")
    outputVarList.pushBack("a")
    sample.setEltAtRank("inputVarList", inputVarList)
    sample.setEltAtRank("outputVarList", outputVarList)
    variable          = pilot.SequenceAny_New(self.runtime.getTypeCode("double"))
    variable_sequence = pilot.SequenceAny_New(variable.getType())
    state_sequence    = pilot.SequenceAny_New(variable_sequence.getType())
    time_sequence     = pilot.SequenceAny_New(state_sequence.getType())
    variable.pushBack(1.0)
    variable_sequence.pushBack(variable)
    state_sequence.pushBack(variable_sequence)
    time_sequence.pushBack(state_sequence)
    sample.setEltAtRank("inputValues", time_sequence)

    # Add observer values in specific parameters
    specificParameters = pilot.SequenceAny_New(self.runtime.getTypeCode("SALOME_TYPES/Parameter"))

    # Switch Value
    obs_switch = pilot.StructAny_New(self.runtime.getTypeCode('SALOME_TYPES/Parameter'))
    obs_switch.setEltAtRank("name", "switch_value")
    obs_switch.setEltAtRank("value", dumps(self.da_study.observers_dict[info]["number"]))
    specificParameters.pushBack(obs_switch)

    # Var
    var_struct = pilot.StructAny_New(self.runtime.getTypeCode('SALOME_TYPES/Parameter'))
    var_struct.setEltAtRank("name", "var")

    # Remove Data Observer, so you can ...
    var.removeDataObserver(self.obs)
    # Pickle then ...
    var_struct.setEltAtRank("value", dumps(var))
    specificParameters.pushBack(var_struct)
    # Add Again Data Observer
    if self.da_study.observers_dict[info]["scheduler"] != "":
      self.ADD.setObserver(Variable = info, ObjectFunction = self.obs, Scheduler = self.da_study.observers_dict[info]["scheduler"], Info = info)
    else:
      self.ADD.setObserver(Variable = info, ObjectFunction = self.obs, Info = info)

    # Info
    info_struct = pilot.StructAny_New(self.runtime.getTypeCode('SALOME_TYPES/Parameter'))
    info_struct.setEltAtRank("name", "info")
    info_struct.setEltAtRank("value", dumps(self.da_study.observers_dict[info]["info"]))
    specificParameters.pushBack(info_struct)

    sample.setEltAtRank("specificParameters", specificParameters)

    self.counter_lock.acquire()
    self.sample_counter += 1
    local_counter = self.sample_counter
    self.pool.pushInSample(local_counter, sample)

    # Wait
    try:
      while True:
        self.signalMasterAndWait()
        if self.isTerminationRequested():
          self.pool.destroyAll()
        else:
          # Get current Id
          sample_id = self.pool.getCurrentId()
          if sample_id == local_counter:
            # 5: Release lock
            # Have to be done before but need a new implementation
            # of the optimizer loop
            self.counter_lock.release()
            break
    except:
      print("Exception in user code:")
      print('-'*60)
      traceback.print_exc(file=sys.stdout)
      print('-'*60)

  def getAlgoResult(self):
#     # Remove data observers, required to pickle assimilation study object
#     for observer_name in list(self.da_study.observers_dict.keys()):
#       self.ADD.removeDataObserver(observer_name, self.obs)
    if sys.version_info.major < 3:
        self.da_study.YI_prepare_to_pickle()
        return pickle.dumps(self.da_study)
    else:
        return pickle.dumps(self.da_study.YI_prepare_to_pickle())

  # Obligatoire ???
  def finish(self):
    pass
  def parseFileToInit(self,fileName):
    pass

  # Fonctions qui ne changent pas
  def setPool(self,pool):
    self.pool=pool
  def getTCForIn(self):
    return self.tin
  def getTCForOut(self):
    return self.tout
  def getTCForAlgoInit(self):
    return self.pyobject
  def getTCForAlgoResult(self):
    return self.pyobject




# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: André Ribes, andre.ribes@edf.fr, EDF R&D

"""
    Ce module sert pour charger les paramètres de configuration d'EFICAS
"""
# Modules Python
# print "passage dans la surcharge de configuration pour Adao"
import os
from PyQt5.QtGui  import *

# Modules Eficas
from InterfaceQT4 import configuration

# Classe de base permettant de lire, afficher
# et sauvegarder les fichiers utilisateurs
class CONFIG(configuration.configBase): # configuration.CONFIG_BASE):

  def __init__(self,appli,repIni):

    # self.labels_eficas = ['lang']
    self.labels_eficas = ['lang','rep_cata','catalogues','closeAutreCommande','closeFrameRechercheCommande','closeEntete','taille']
    configuration.configBase.__init__(self,appli,repIni)
    #~ configuration.CONFIG_BASE.__init__(self,appli,repIni)

    self.rep_user         = os.environ["HOME"]
    self.appli            = appli
    self.code             = appli.code
    # self.lang           = "fr"
    self.rep_ini          = repIni
    self.rep_mat          = " " # Compatbilite Aster
    self.savedir          = self.rep_user
    self.generator_module = "generator_adao"
    self.convert_module   = "convert_adao"

    # Format des catalogues...
    # (code, version, catalogue, formatIn, formatOut)
    # Il faut les mettre dans un tuple
    self.catalogues = (("ADAO", "V0", os.path.join(self.rep_ini, 'ADAO_Cata_V0.py'), "adao"),)
# JPA 21/09/2017 : self.catalogues = (("ADAO", "V0", os.path.join(self.rep_ini, 'ADAO_Cata_V0.py'), "python", "adao"),)

def make_config(appli,rep):

    return CONFIG(appli,rep)

def make_config_style(appli,rep):

    return None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import numpy
from daCore import BasicObjects, NumericObjects
from daAlgorithms.Atoms import mmqr

# ==============================================================================
class ElementaryAlgorithm(BasicObjects.Algorithm):
    def __init__(self):
        BasicObjects.Algorithm.__init__(self, "QUANTILEREGRESSION")
        self.defineRequiredParameter(
            name     = "Quantile",
            default  = 0.5,
            typecast = float,
            message  = "Quantile pour la regression de quantile",
            minval   = 0.,
            maxval   = 1.,
            )
        self.defineRequiredParameter(
            name     = "Minimizer",
            default  = "MMQR",
            typecast = str,
            message  = "Minimiseur utilisé",
            listval  = ["MMQR",],
            )
        self.defineRequiredParameter(
            name     = "MaximumNumberOfIterations",
            default  = 15000,
            typecast = int,
            message  = "Nombre maximal de pas d'optimisation",
            minval   = 1,
            oldname  = "MaximumNumberOfSteps",
            )
        self.defineRequiredParameter(
            name     = "CostDecrementTolerance",
            default  = 1.e-6,
            typecast = float,
            message  = "Maximum de variation de la fonction d'estimation lors de l'arrêt",
            )
        self.defineRequiredParameter(
            name     = "StoreInternalVariables",
            default  = False,
            typecast = bool,
            message  = "Stockage des variables internes ou intermédiaires du calcul",
            )
        self.defineRequiredParameter(
            name     = "StoreSupplementaryCalculations",
            default  = [],
            typecast = tuple,
            message  = "Liste de calculs supplémentaires à stocker et/ou effectuer",
            listval  = [
                "Analysis",
                "BMA",
                "CostFunctionJ",
                "CostFunctionJb",
                "CostFunctionJo",
                "CurrentIterationNumber",
                "CurrentState",
                "Innovation",
                "OMA",
                "OMB",
                "SimulatedObservationAtBackground",
                "SimulatedObservationAtCurrentState",
                "SimulatedObservationAtOptimum",
                ]
            )
        self.defineRequiredParameter( # Pas de type
            name     = "Bounds",
            message  = "Liste des valeurs de bornes",
            )
        self.defineRequiredParameter(
            name     = "InitializationPoint",
            typecast = numpy.ravel,
            message  = "État initial imposé (par défaut, c'est l'ébauche si None)",
            )
        self.requireInputArguments(
            mandatory= ("Xb", "Y", "HO" ),
            )
        self.setAttributes(tags=(
            "Optimization",
            "Risk",
            "Variational",
            ))

    def run(self, Xb=None, Y=None, U=None, HO=None, EM=None, CM=None, R=None, B=None, Q=None, Parameters=None):
        self._pre_run(Parameters, Xb, Y, U, HO, EM, CM, R, B, Q)
        self._parameters["Bounds"] = NumericObjects.ForceNumericBounds( self._parameters["Bounds"] )
        #
        Hm = HO["Direct"].appliedTo
        #
        def CostFunction(x):
            _X = numpy.asarray(x).reshape((-1,1))
            if self._parameters["StoreInternalVariables"] or \
                self._toStore("CurrentState"):
                self.StoredVariables["CurrentState"].store( _X )
            _HX = numpy.asarray(Hm( _X )).reshape((-1,1))
            if self._toStore("SimulatedObservationAtCurrentState"):
                self.StoredVariables["SimulatedObservationAtCurrentState"].store( _HX )
            Jb  = 0.
            Jo  = 0.
            J   = Jb + Jo
            #
            self.StoredVariables["CurrentIterationNumber"].store( len(self.StoredVariables["CostFunctionJ"]) )
            self.StoredVariables["CostFunctionJb"].store( Jb )
            self.StoredVariables["CostFunctionJo"].store( Jo )
            self.StoredVariables["CostFunctionJ" ].store( J )
            return _HX
        #
        def GradientOfCostFunction(x):
            _X = numpy.asarray(x).reshape((-1,1))
            Hg = HO["Tangent"].asMatrix( _X )
            return Hg
        #
        Xini = self._parameters["InitializationPoint"]
        #
        # Minimisation de la fonctionnelle
        # --------------------------------
        if self._parameters["Minimizer"] == "MMQR":
            Minimum, J_optimal, Informations = mmqr.mmqr(
                func        = CostFunction,
                x0          = Xini,
                fprime      = GradientOfCostFunction,
                bounds      = self._parameters["Bounds"],
                quantile    = self._parameters["Quantile"],
                maxfun      = self._parameters["MaximumNumberOfIterations"],
                toler       = self._parameters["CostDecrementTolerance"],
                y           = Y,
                )
        else:
            raise ValueError("Error in minimizer name: %s is unkown"%self._parameters["Minimizer"])
        #
        # Obtention de l'analyse
        # ----------------------
        Xa = Minimum
        #
        self.StoredVariables["Analysis"].store( Xa )
        #
        # Calculs et/ou stockages supplémentaires
        # ---------------------------------------
        if self._toStore("OMA") or \
            self._toStore("SimulatedObservationAtOptimum"):
            HXa = Hm(Xa).reshape((-1,1))
        if self._toStore("Innovation") or \
            self._toStore("OMB") or \
            self._toStore("SimulatedObservationAtBackground"):
            HXb = Hm(Xb).reshape((-1,1))
            Innovation = Y - HXb
        if self._toStore("Innovation"):
            self.StoredVariables["Innovation"].store( Innovation )
        if self._toStore("OMB"):
            self.StoredVariables["OMB"].store( Innovation )
        if self._toStore("BMA"):
            self.StoredVariables["BMA"].store( numpy.ravel(Xb) - numpy.ravel(Xa) )
        if self._toStore("OMA"):
            self.StoredVariables["OMA"].store( Y - HXa )
        if self._toStore("SimulatedObservationAtBackground"):
            self.StoredVariables["SimulatedObservationAtBackground"].store( HXb )
        if self._toStore("SimulatedObservationAtOptimum"):
            self.StoredVariables["SimulatedObservationAtOptimum"].store( HXa )
        #
        self._post_run(HO)
        return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: André Ribes, andre.ribes@edf.fr, EDF R&D

import ADAO_COMPONENT__POA
import SALOME_ComponentPy
import SALOME_DriverPy
import SALOME_Embedded_NamingService_ClientPy

from daUtils.adaoLogger import *

class ADAO(ADAO_COMPONENT__POA.ADAO_ENGINE,
           SALOME_ComponentPy.SALOME_ComponentPy_i,
           SALOME_DriverPy.SALOME_DriverPy_i):
  """
      Pour etre un composant SALOME cette classe Python
      doit avoir le nom du composant et heriter de la
      classe ADAO_COMPONENT issue de la compilation de l'idl
      par omniidl et de la classe SALOME_ComponentPy_i
      qui porte les services generaux d'un composant SALOME
  """
  def __init__ ( self, orb, poa, contID, containerName, instanceName,
      interfaceName ):
    debug("Creating ADAO component instance", "ENGINE")
    SALOME_ComponentPy.SALOME_ComponentPy_i.__init__(self, orb, poa,
        contID, containerName, instanceName, interfaceName)
    SALOME_DriverPy.SALOME_DriverPy_i.__init__(self, interfaceName)

    # On stocke dans l'attribut _naming_service, une ref sur le Naming Service
    #
    emb_ns = self._contId.get_embedded_NS_if_ssl()
    import CORBA
    if CORBA.is_nil(emb_ns):
        self._naming_service = SALOME_ComponentPy.SALOME_NamingServicePy_i( self._orb )
    else:
        self._naming_service = SALOME_Embedded_NamingService_ClientPy.SALOME_Embedded_NamingService_ClientPy(emb_ns)
    #

  def print_ping():
    info("ADAO ENGINE Ping", "ENGINE")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import numpy
from daCore import BasicObjects
from daAlgorithms.Atoms import enks, etkf, ienkf, mlef, senkf

# ==============================================================================
class ElementaryAlgorithm(BasicObjects.Algorithm):
    def __init__(self):
        BasicObjects.Algorithm.__init__(self, "ENSEMBLEKALMANFILTER")
        self.defineRequiredParameter(
            name     = "Variant",
            default  = "EnKF",
            typecast = str,
            message  = "Variant ou formulation de la méthode",
            listval  = [
                "EnKF",
                "ETKF",
                "ETKF-N",
                "MLEF",
                "IEnKF",
                "EnKS",
                ],
            listadv  = [
                "StochasticEnKF",
                "EnKF-05",
                "EnKF-16",
                "ETKF-KFF",
                "ETKF-VAR",
                "ETKF-N-11",
                "ETKF-N-15",
                "ETKF-N-16",
                "MLEF-T",
                "MLEF-B",
                "IEnKF-T",
                "IEnKF-B",
                "EnKS-KFF",
                "IEKF",
                "E3DVAR",
                "E3DVAR-EnKF",
                "E3DVAR-ETKF",
                "E3DVAR-MLEF",
                ],
            )
        self.defineRequiredParameter(
            name     = "NumberOfMembers",
            default  = 100,
            typecast = int,
            message  = "Nombre de membres dans l'ensemble",
            minval   = 2,
            )
        self.defineRequiredParameter(
            name     = "EstimationOf",
            default  = "State",
            typecast = str,
            message  = "Estimation d'etat ou de parametres",
            listval  = ["State", "Parameters"],
            )
        self.defineRequiredParameter(
            name     = "InflationType",
            default  = "MultiplicativeOnAnalysisAnomalies",
            typecast = str,
            message  = "Méthode d'inflation d'ensemble",
            listval  = [
                "MultiplicativeOnAnalysisAnomalies",
                "MultiplicativeOnBackgroundAnomalies",
                ],
            listadv  = [
                "MultiplicativeOnAnalysisCovariance",
                "MultiplicativeOnBackgroundCovariance",
                "AdditiveOnAnalysisCovariance",
                "AdditiveOnBackgroundCovariance",
                "HybridOnBackgroundCovariance",
                "Relaxation",
                ],
            )
        self.defineRequiredParameter(
            name     = "InflationFactor",
            default  = 1.,
            typecast = float,
            message  = "Facteur d'inflation",
            minval   = 0.,
            )
        self.defineRequiredParameter(
            name     = "SmootherLagL",
            default  = 0,
            typecast = int,
            message  = "Nombre d'intervalles de temps de lissage dans le passé",
            minval   = 0,
            )
        self.defineRequiredParameter(
            name     = "HybridCovarianceEquilibrium",
            default  = 0.5,
            typecast = float,
            message  = "Facteur d'équilibre entre la covariance statique et la covariance d'ensemble en hybride variationnel",
            minval   = 0.,
            maxval   = 1.,
            )
        self.defineRequiredParameter(
            name     = "HybridMaximumNumberOfIterations",
            default  = 15000,
            typecast = int,
            message  = "Nombre maximal de pas d'optimisation en hybride variationnel",
            minval   = -1,
            )
        self.defineRequiredParameter(
            name     = "HybridCostDecrementTolerance",
            default  = 1.e-7,
            typecast = float,
            message  = "Diminution relative minimale du coût lors de l'arrêt en hybride variationnel",
            minval   = 0.,
            )
        self.defineRequiredParameter(
            name     = "SetSeed",
            typecast = numpy.random.seed,
            message  = "Graine fixée pour le générateur aléatoire",
            )
        self.defineRequiredParameter(
            name     = "StoreInternalVariables",
            default  = False,
            typecast = bool,
            message  = "Stockage des variables internes ou intermédiaires du calcul",
            )
        self.defineRequiredParameter(
            name     = "StoreSupplementaryCalculations",
            default  = [],
            typecast = tuple,
            message  = "Liste de calculs supplémentaires à stocker et/ou effectuer",
            listval  = [
                "Analysis",
                "APosterioriCorrelations",
                "APosterioriCovariance",
                "APosterioriStandardDeviations",
                "APosterioriVariances",
                "BMA",
                "CostFunctionJ",
                "CostFunctionJAtCurrentOptimum",
                "CostFunctionJb",
                "CostFunctionJbAtCurrentOptimum",
                "CostFunctionJo",
                "CostFunctionJoAtCurrentOptimum",
                "CurrentIterationNumber",
                "CurrentOptimum",
                "CurrentState",
                "ForecastCovariance",
                "ForecastState",
                "IndexOfOptimum",
                "InnovationAtCurrentAnalysis",
                "InnovationAtCurrentState",
                "SimulatedObservationAtCurrentAnalysis",
                "SimulatedObservationAtCurrentOptimum",
                "SimulatedObservationAtCurrentState",
                ],
            listadv  = [
                "CurrentEnsembleState",
                ],
            )
        self.requireInputArguments(
            mandatory= ("Xb", "Y", "HO", "R", "B"),
            optional = ("U", "EM", "CM", "Q"),
            )
        self.setAttributes(tags=(
            "DataAssimilation",
            "NonLinear",
            "Filter",
            "Ensemble",
            "Dynamic",
            ))

    def run(self, Xb=None, Y=None, U=None, HO=None, EM=None, CM=None, R=None, B=None, Q=None, Parameters=None):
        self._pre_run(Parameters, Xb, Y, U, HO, EM, CM, R, B, Q)
        #
        #--------------------------
        # Default EnKF = EnKF-16 = StochasticEnKF
        if   self._parameters["Variant"] == "EnKF-05":
            senkf.senkf(self, Xb, Y, U, HO, EM, CM, R, B, Q, VariantM="KalmanFilterFormula05")
        #
        elif self._parameters["Variant"] in ["EnKF-16", "StochasticEnKF", "EnKF"]:
            senkf.senkf(self, Xb, Y, U, HO, EM, CM, R, B, Q, VariantM="KalmanFilterFormula16")
        #
        #--------------------------
        # Default ETKF = ETKF-KFF
        elif self._parameters["Variant"] in ["ETKF-KFF", "ETKF"]:
            etkf.etkf(self, Xb, Y, U, HO, EM, CM, R, B, Q, VariantM="KalmanFilterFormula")
        #
        elif self._parameters["Variant"] == "ETKF-VAR":
            etkf.etkf(self, Xb, Y, U, HO, EM, CM, R, B, Q, VariantM="Variational")
        #
        #--------------------------
        # Default ETKF-N = ETKF-N-16
        elif self._parameters["Variant"] == "ETKF-N-11":
            etkf.etkf(self, Xb, Y, U, HO, EM, CM, R, B, Q, VariantM="FiniteSize11")
        #
        elif self._parameters["Variant"] == "ETKF-N-15":
            etkf.etkf(self, Xb, Y, U, HO, EM, CM, R, B, Q, VariantM="FiniteSize15")
        #
        elif self._parameters["Variant"] in ["ETKF-N-16", "ETKF-N"]:
            etkf.etkf(self, Xb, Y, U, HO, EM, CM, R, B, Q, VariantM="FiniteSize16")
        #
        #--------------------------
        # Default MLEF = MLEF-T
        elif self._parameters["Variant"] in ["MLEF-T", "MLEF"]:
            mlef.mlef(self, Xb, Y, U, HO, EM, CM, R, B, Q, BnotT=False)
        #
        elif self._parameters["Variant"] == "MLEF-B":
            mlef.mlef(self, Xb, Y, U, HO, EM, CM, R, B, Q, BnotT=True)
        #
        #--------------------------
        # Default IEnKF = IEnKF-T
        elif self._parameters["Variant"] in ["IEnKF-T", "IEnKF"]:
            ienkf.ienkf(self, Xb, Y, U, HO, EM, CM, R, B, Q, BnotT=False)
        #
        elif self._parameters["Variant"] in ["IEnKF-B", "IEKF"]:
            ienkf.ienkf(self, Xb, Y, U, HO, EM, CM, R, B, Q, BnotT=True)
        #
        #--------------------------
        # Default EnKS = EnKS-KFF
        elif self._parameters["Variant"] in ["EnKS-KFF", "EnKS"]:
            enks.enks(self, Xb, Y, U, HO, EM, CM, R, B, Q, VariantM="EnKS16-KalmanFilterFormula")
        #
        #--------------------------
        # Default E3DVAR = E3DVAR-ETKF
        elif self._parameters["Variant"] == "E3DVAR-EnKF":
            senkf.senkf(self, Xb, Y, U, HO, EM, CM, R, B, Q, Hybrid="E3DVAR")
        #
        elif self._parameters["Variant"] in ["E3DVAR-ETKF", "E3DVAR"]:
            etkf.etkf(self, Xb, Y, U, HO, EM, CM, R, B, Q, Hybrid="E3DVAR")
        #
        elif self._parameters["Variant"] == "E3DVAR-MLEF":
            mlef.mlef(self, Xb, Y, U, HO, EM, CM, R, B, Q, Hybrid="E3DVAR")
        #
        #--------------------------
        else:
            raise ValueError("Error in Variant name: %s"%self._parameters["Variant"])
        #
        self._post_run(HO)
        return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import numpy, logging, scipy.optimize
from daCore import BasicObjects, PlatformInfo

# ==============================================================================
class ElementaryAlgorithm(BasicObjects.Algorithm):
    def __init__(self):
        BasicObjects.Algorithm.__init__(self, "DERIVATIVEFREEOPTIMIZATION")
        self.defineRequiredParameter(
            name     = "Minimizer",
            default  = "BOBYQA",
            typecast = str,
            message  = "Minimiseur utilisé",
            listval  = [
                "BOBYQA",
                "COBYLA",
                "NEWUOA",
                "POWELL",
                "SIMPLEX",
                "SUBPLEX",
                ],
            )
        self.defineRequiredParameter(
            name     = "MaximumNumberOfIterations",
            default  = 15000,
            typecast = int,
            message  = "Nombre maximal de pas d'optimisation",
            minval   = -1,
            oldname  = "MaximumNumberOfSteps",
            )
        self.defineRequiredParameter(
            name     = "MaximumNumberOfFunctionEvaluations",
            default  = 15000,
            typecast = int,
            message  = "Nombre maximal d'évaluations de la fonction",
            minval   = -1,
            )
        self.defineRequiredParameter(
            name     = "StateVariationTolerance",
            default  = 1.e-4,
            typecast = float,
            message  = "Variation relative maximale de l'état lors de l'arrêt",
            )
        self.defineRequiredParameter(
            name     = "CostDecrementTolerance",
            default  = 1.e-7,
            typecast = float,
            message  = "Diminution relative minimale du cout lors de l'arrêt",
            )
        self.defineRequiredParameter(
            name     = "QualityCriterion",
            default  = "AugmentedWeightedLeastSquares",
            typecast = str,
            message  = "Critère de qualité utilisé",
            listval  = [
                "AugmentedWeightedLeastSquares", "AWLS", "DA",
                "WeightedLeastSquares", "WLS",
                "LeastSquares", "LS", "L2",
                "AbsoluteValue", "L1",
                "MaximumError", "ME",
                ],
            )
        self.defineRequiredParameter(
            name     = "StoreInternalVariables",
            default  = False,
            typecast = bool,
            message  = "Stockage des variables internes ou intermédiaires du calcul",
            )
        self.defineRequiredParameter(
            name     = "StoreSupplementaryCalculations",
            default  = [],
            typecast = tuple,
            message  = "Liste de calculs supplémentaires à stocker et/ou effectuer",
            listval  = [
                "Analysis",
                "BMA",
                "CostFunctionJ",
                "CostFunctionJb",
                "CostFunctionJo",
                "CostFunctionJAtCurrentOptimum",
                "CostFunctionJbAtCurrentOptimum",
                "CostFunctionJoAtCurrentOptimum",
                "CurrentIterationNumber",
                "CurrentOptimum",
                "CurrentState",
                "IndexOfOptimum",
                "Innovation",
                "InnovationAtCurrentState",
                "OMA",
                "OMB",
                "SimulatedObservationAtBackground",
                "SimulatedObservationAtCurrentOptimum",
                "SimulatedObservationAtCurrentState",
                "SimulatedObservationAtOptimum",
                ]
            )
        self.defineRequiredParameter( # Pas de type
            name     = "Bounds",
            message  = "Liste des valeurs de bornes",
            )
        self.requireInputArguments(
            mandatory= ("Xb", "Y", "HO", "R", "B"),
            )
        self.setAttributes(tags=(
            "Optimization",
            "NonLinear",
            "MetaHeuristic",
            ))

    def run(self, Xb=None, Y=None, U=None, HO=None, EM=None, CM=None, R=None, B=None, Q=None, Parameters=None):
        self._pre_run(Parameters, Xb, Y, U, HO, EM, CM, R, B, Q)
        #
        if not PlatformInfo.has_nlopt and not self._parameters["Minimizer"] in ["COBYLA", "POWELL", "SIMPLEX"]:
            logging.warning("%s Minimization by SIMPLEX is forced because %s is unavailable (COBYLA, POWELL are also available)"%(self._name,self._parameters["Minimizer"]))
            self._parameters["Minimizer"] = "SIMPLEX"
        #
        Hm = HO["Direct"].appliedTo
        #
        BI = B.getI()
        RI = R.getI()
        #
        def CostFunction(x, QualityMeasure="AugmentedWeightedLeastSquares"):
            _X  = numpy.ravel( x ).reshape((-1,1))
            _HX = numpy.ravel( Hm( _X ) ).reshape((-1,1))
            _Innovation = Y - _HX
            self.StoredVariables["CurrentState"].store( _X )
            if self._toStore("SimulatedObservationAtCurrentState") or \
                self._toStore("SimulatedObservationAtCurrentOptimum"):
                self.StoredVariables["SimulatedObservationAtCurrentState"].store( _HX )
            if self._toStore("InnovationAtCurrentState"):
                self.StoredVariables["InnovationAtCurrentState"].store( _Innovation )
            #
            if QualityMeasure in ["AugmentedWeightedLeastSquares","AWLS","DA"]:
                if BI is None or RI is None:
                    raise ValueError("Background and Observation error covariance matrices has to be properly defined!")
                Jb  = 0.5 * (_X - Xb).T @ (BI @ (_X - Xb))
                Jo  = 0.5 * _Innovation.T @ (RI @ _Innovation)
            elif QualityMeasure in ["WeightedLeastSquares","WLS"]:
                if RI is None:
                    raise ValueError("Observation error covariance matrix has to be properly defined!")
                Jb  = 0.
                Jo  = 0.5 * _Innovation.T @ (RI @ _Innovation)
            elif QualityMeasure in ["LeastSquares","LS","L2"]:
                Jb  = 0.
                Jo  = 0.5 * _Innovation.T @ _Innovation
            elif QualityMeasure in ["AbsoluteValue","L1"]:
                Jb  = 0.
                Jo  = numpy.sum( numpy.abs(_Innovation) )
            elif QualityMeasure in ["MaximumError","ME"]:
                Jb  = 0.
                Jo  = numpy.max( numpy.abs(_Innovation) )
            #
            J   = float( Jb ) + float( Jo )
            #
            self.StoredVariables["CurrentIterationNumber"].store( len(self.StoredVariables["CostFunctionJ"]) )
            self.StoredVariables["CostFunctionJb"].store( Jb )
            self.StoredVariables["CostFunctionJo"].store( Jo )
            self.StoredVariables["CostFunctionJ" ].store( J )
            if self._toStore("IndexOfOptimum") or \
                self._toStore("CurrentOptimum") or \
                self._toStore("CostFunctionJAtCurrentOptimum") or \
                self._toStore("CostFunctionJbAtCurrentOptimum") or \
                self._toStore("CostFunctionJoAtCurrentOptimum") or \
                self._toStore("SimulatedObservationAtCurrentOptimum"):
                IndexMin = numpy.argmin( self.StoredVariables["CostFunctionJ"][nbPreviousSteps:] ) + nbPreviousSteps
            if self._toStore("IndexOfOptimum"):
                self.StoredVariables["IndexOfOptimum"].store( IndexMin )
            if self._toStore("CurrentOptimum"):
                self.StoredVariables["CurrentOptimum"].store( self.StoredVariables["CurrentState"][IndexMin] )
            if self._toStore("SimulatedObservationAtCurrentOptimum"):
                self.StoredVariables["SimulatedObservationAtCurrentOptimum"].store( self.StoredVariables["SimulatedObservationAtCurrentState"][IndexMin] )
            if self._toStore("CostFunctionJAtCurrentOptimum"):
                self.StoredVariables["CostFunctionJAtCurrentOptimum" ].store( self.StoredVariables["CostFunctionJ" ][IndexMin] )
            if self._toStore("CostFunctionJbAtCurrentOptimum"):
                self.StoredVariables["CostFunctionJbAtCurrentOptimum"].store( self.StoredVariables["CostFunctionJb"][IndexMin] )
            if self._toStore("CostFunctionJoAtCurrentOptimum"):
                self.StoredVariables["CostFunctionJoAtCurrentOptimum"].store( self.StoredVariables["CostFunctionJo"][IndexMin] )
            return J
        #
        Xini = numpy.ravel(Xb)
        if len(Xini) < 2 and self._parameters["Minimizer"] == "NEWUOA":
            raise ValueError("The minimizer %s can not be used when the optimisation state dimension is 1. Please choose another minimizer."%self._parameters["Minimizer"])
        #
        # Minimisation de la fonctionnelle
        # --------------------------------
        nbPreviousSteps = self.StoredVariables["CostFunctionJ"].stepnumber()
        #
        if self._parameters["Minimizer"] == "POWELL":
            Minimum, J_optimal, direc, niter, nfeval, rc = scipy.optimize.fmin_powell(
                func        = CostFunction,
                x0          = Xini,
                args        = (self._parameters["QualityCriterion"],),
                maxiter     = self._parameters["MaximumNumberOfIterations"]-1,
                maxfun      = self._parameters["MaximumNumberOfFunctionEvaluations"],
                xtol        = self._parameters["StateVariationTolerance"],
                ftol        = self._parameters["CostDecrementTolerance"],
                full_output = True,
                disp        = self._parameters["optdisp"],
                )
        elif self._parameters["Minimizer"] == "COBYLA" and not PlatformInfo.has_nlopt:
            def make_constraints(bounds):
                constraints = []
                for (i,(a,b)) in enumerate(bounds):
                    lower = lambda x: x[i] - a
                    upper = lambda x: b - x[i]
                    constraints = constraints + [lower] + [upper]
                return constraints
            if self._parameters["Bounds"] is None:
                raise ValueError("Bounds have to be given for all axes as a list of lower/upper pairs!")
            Minimum = scipy.optimize.fmin_cobyla(
                func        = CostFunction,
                x0          = Xini,
                cons        = make_constraints( self._parameters["Bounds"] ),
                args        = (self._parameters["QualityCriterion"],),
                consargs    = (), # To avoid extra-args
                maxfun      = self._parameters["MaximumNumberOfFunctionEvaluations"],
                rhobeg      = 1.0,
                rhoend      = self._parameters["StateVariationTolerance"],
                catol       = 2.*self._parameters["StateVariationTolerance"],
                disp        = self._parameters["optdisp"],
                )
        elif self._parameters["Minimizer"] == "COBYLA" and PlatformInfo.has_nlopt:
            import nlopt
            opt = nlopt.opt(nlopt.LN_COBYLA, Xini.size)
            def _f(_Xx, Grad):
                # DFO, so no gradient
                return CostFunction(_Xx, self._parameters["QualityCriterion"])
            opt.set_min_objective(_f)
            if self._parameters["Bounds"] is not None:
                lub = numpy.array(self._parameters["Bounds"],dtype=float).reshape((Xini.size,2))
                lb = lub[:,0] ; lb[numpy.isnan(lb)] = -float('inf')
                ub = lub[:,1] ; ub[numpy.isnan(ub)] = +float('inf')
                if self._parameters["optdisp"]:
                    print("%s: upper bounds %s"%(opt.get_algorithm_name(),ub))
                    print("%s: lower bounds %s"%(opt.get_algorithm_name(),lb))
                opt.set_upper_bounds(ub)
                opt.set_lower_bounds(lb)
            opt.set_ftol_rel(self._parameters["CostDecrementTolerance"])
            opt.set_xtol_rel(2.*self._parameters["StateVariationTolerance"])
            opt.set_maxeval(self._parameters["MaximumNumberOfFunctionEvaluations"])
            Minimum = opt.optimize( Xini )
            if self._parameters["optdisp"]:
                print("%s: optimal state: %s"%(opt.get_algorithm_name(),Minimum))
                print("%s: minimum of J: %s"%(opt.get_algorithm_name(),opt.last_optimum_value()))
                print("%s: return code: %i"%(opt.get_algorithm_name(),opt.last_optimize_result()))
        elif self._parameters["Minimizer"] == "SIMPLEX" and not PlatformInfo.has_nlopt:
            Minimum, J_optimal, niter, nfeval, rc = scipy.optimize.fmin(
                func        = CostFunction,
                x0          = Xini,
                args        = (self._parameters["QualityCriterion"],),
                maxiter     = self._parameters["MaximumNumberOfIterations"]-1,
                maxfun      = self._parameters["MaximumNumberOfFunctionEvaluations"],
                xtol        = self._parameters["StateVariationTolerance"],
                ftol        = self._parameters["CostDecrementTolerance"],
                full_output = True,
                disp        = self._parameters["optdisp"],
                )
        elif self._parameters["Minimizer"] == "SIMPLEX" and PlatformInfo.has_nlopt:
            import nlopt
            opt = nlopt.opt(nlopt.LN_NELDERMEAD, Xini.size)
            def _f(_Xx, Grad):
                # DFO, so no gradient
                return CostFunction(_Xx, self._parameters["QualityCriterion"])
            opt.set_min_objective(_f)
            if self._parameters["Bounds"] is not None:
                lub = numpy.array(self._parameters["Bounds"],dtype=float).reshape((Xini.size,2))
                lb = lub[:,0] ; lb[numpy.isnan(lb)] = -float('inf')
                ub = lub[:,1] ; ub[numpy.isnan(ub)] = +float('inf')
                if self._parameters["optdisp"]:
                    print("%s: upper bounds %s"%(opt.get_algorithm_name(),ub))
                    print("%s: lower bounds %s"%(opt.get_algorithm_name(),lb))
                opt.set_upper_bounds(ub)
                opt.set_lower_bounds(lb)
            opt.set_ftol_rel(self._parameters["CostDecrementTolerance"])
            opt.set_xtol_rel(2.*self._parameters["StateVariationTolerance"])
            opt.set_maxeval(self._parameters["MaximumNumberOfFunctionEvaluations"])
            Minimum = opt.optimize( Xini )
            if self._parameters["optdisp"]:
                print("%s: optimal state: %s"%(opt.get_algorithm_name(),Minimum))
                print("%s: minimum of J: %s"%(opt.get_algorithm_name(),opt.last_optimum_value()))
                print("%s: return code: %i"%(opt.get_algorithm_name(),opt.last_optimize_result()))
        elif self._parameters["Minimizer"] == "BOBYQA" and PlatformInfo.has_nlopt:
            import nlopt
            opt = nlopt.opt(nlopt.LN_BOBYQA, Xini.size)
            def _f(_Xx, Grad):
                # DFO, so no gradient
                return CostFunction(_Xx, self._parameters["QualityCriterion"])
            opt.set_min_objective(_f)
            if self._parameters["Bounds"] is not None:
                lub = numpy.array(self._parameters["Bounds"],dtype=float).reshape((Xini.size,2))
                lb = lub[:,0] ; lb[numpy.isnan(lb)] = -float('inf')
                ub = lub[:,1] ; ub[numpy.isnan(ub)] = +float('inf')
                if self._parameters["optdisp"]:
                    print("%s: upper bounds %s"%(opt.get_algorithm_name(),ub))
                    print("%s: lower bounds %s"%(opt.get_algorithm_name(),lb))
                opt.set_upper_bounds(ub)
                opt.set_lower_bounds(lb)
            opt.set_ftol_rel(self._parameters["CostDecrementTolerance"])
            opt.set_xtol_rel(2.*self._parameters["StateVariationTolerance"])
            opt.set_maxeval(self._parameters["MaximumNumberOfFunctionEvaluations"])
            Minimum = opt.optimize( Xini )
            if self._parameters["optdisp"]:
                print("%s: optimal state: %s"%(opt.get_algorithm_name(),Minimum))
                print("%s: minimum of J: %s"%(opt.get_algorithm_name(),opt.last_optimum_value()))
                print("%s: return code: %i"%(opt.get_algorithm_name(),opt.last_optimize_result()))
        elif self._parameters["Minimizer"] == "NEWUOA" and PlatformInfo.has_nlopt:
            import nlopt
            opt = nlopt.opt(nlopt.LN_NEWUOA, Xini.size)
            def _f(_Xx, Grad):
                # DFO, so no gradient
                return CostFunction(_Xx, self._parameters["QualityCriterion"])
            opt.set_min_objective(_f)
            if self._parameters["Bounds"] is not None:
                lub = numpy.array(self._parameters["Bounds"],dtype=float).reshape((Xini.size,2))
                lb = lub[:,0] ; lb[numpy.isnan(lb)] = -float('inf')
                ub = lub[:,1] ; ub[numpy.isnan(ub)] = +float('inf')
                if self._parameters["optdisp"]:
                    print("%s: upper bounds %s"%(opt.get_algorithm_name(),ub))
                    print("%s: lower bounds %s"%(opt.get_algorithm_name(),lb))
                opt.set_upper_bounds(ub)
                opt.set_lower_bounds(lb)
            opt.set_ftol_rel(self._parameters["CostDecrementTolerance"])
            opt.set_xtol_rel(2.*self._parameters["StateVariationTolerance"])
            opt.set_maxeval(self._parameters["MaximumNumberOfFunctionEvaluations"])
            Minimum = opt.optimize( Xini )
            if self._parameters["optdisp"]:
                print("%s: optimal state: %s"%(opt.get_algorithm_name(),Minimum))
                print("%s: minimum of J: %s"%(opt.get_algorithm_name(),opt.last_optimum_value()))
                print("%s: return code: %i"%(opt.get_algorithm_name(),opt.last_optimize_result()))
        elif self._parameters["Minimizer"] == "SUBPLEX" and PlatformInfo.has_nlopt:
            import nlopt
            opt = nlopt.opt(nlopt.LN_SBPLX, Xini.size)
            def _f(_Xx, Grad):
                # DFO, so no gradient
                return CostFunction(_Xx, self._parameters["QualityCriterion"])
            opt.set_min_objective(_f)
            if self._parameters["Bounds"] is not None:
                lub = numpy.array(self._parameters["Bounds"],dtype=float).reshape((Xini.size,2))
                lb = lub[:,0] ; lb[numpy.isnan(lb)] = -float('inf')
                ub = lub[:,1] ; ub[numpy.isnan(ub)] = +float('inf')
                if self._parameters["optdisp"]:
                    print("%s: upper bounds %s"%(opt.get_algorithm_name(),ub))
                    print("%s: lower bounds %s"%(opt.get_algorithm_name(),lb))
                opt.set_upper_bounds(ub)
                opt.set_lower_bounds(lb)
            opt.set_ftol_rel(self._parameters["CostDecrementTolerance"])
            opt.set_xtol_rel(2.*self._parameters["StateVariationTolerance"])
            opt.set_maxeval(self._parameters["MaximumNumberOfFunctionEvaluations"])
            Minimum = opt.optimize( Xini )
            if self._parameters["optdisp"]:
                print("%s: optimal state: %s"%(opt.get_algorithm_name(),Minimum))
                print("%s: minimum of J: %s"%(opt.get_algorithm_name(),opt.last_optimum_value()))
                print("%s: return code: %i"%(opt.get_algorithm_name(),opt.last_optimize_result()))
        else:
            raise ValueError("Error in minimizer name: %s is unkown"%self._parameters["Minimizer"])
        #
        IndexMin = numpy.argmin( self.StoredVariables["CostFunctionJ"][nbPreviousSteps:] ) + nbPreviousSteps
        MinJ     = self.StoredVariables["CostFunctionJ"][IndexMin]
        Minimum  = self.StoredVariables["CurrentState"][IndexMin]
        #
        # Obtention de l'analyse
        # ----------------------
        Xa = Minimum
        #
        self.StoredVariables["Analysis"].store( Xa )
        #
        # Calculs et/ou stockages supplémentaires
        # ---------------------------------------
        if self._toStore("OMA") or \
            self._toStore("SimulatedObservationAtOptimum"):
            if self._toStore("SimulatedObservationAtCurrentState"):
                HXa = self.StoredVariables["SimulatedObservationAtCurrentState"][IndexMin]
            elif self._toStore("SimulatedObservationAtCurrentOptimum"):
                HXa = self.StoredVariables["SimulatedObservationAtCurrentOptimum"][-1]
            else:
                HXa = Hm(Xa)
            HXa = HXa.reshape((-1,1))
        if self._toStore("Innovation") or \
            self._toStore("OMB") or \
            self._toStore("SimulatedObservationAtBackground"):
            HXb = Hm(Xb).reshape((-1,1))
            Innovation = Y - HXb
        if self._toStore("Innovation"):
            self.StoredVariables["Innovation"].store( Innovation )
        if self._toStore("OMB"):
            self.StoredVariables["OMB"].store( Innovation )
        if self._toStore("BMA"):
            self.StoredVariables["BMA"].store( numpy.ravel(Xb) - numpy.ravel(Xa) )
        if self._toStore("OMA"):
            self.StoredVariables["OMA"].store( Y - HXa )
        if self._toStore("SimulatedObservationAtBackground"):
            self.StoredVariables["SimulatedObservationAtBackground"].store( HXb )
        if self._toStore("SimulatedObservationAtOptimum"):
            self.StoredVariables["SimulatedObservationAtOptimum"].store( HXa )
        #
        self._post_run(HO)
        return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



#-*-coding:utf-8-*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

from daCore.Aidsm import Aidsm
#from daCore import Logging
import logging

class daError(Exception):
  def __init__(self, value):
    self.value = value
  def __str__(self):
    return repr(self.value)

class daStudy:

  def __init__(self, name, algorithm, debug):

    self.ADD = Aidsm(name)
    self.algorithm = algorithm
    self.algorithm_dict = None
    self.Background = None
    self.CheckingPoint = None
    self.InputVariables = {}
    self.OutputVariables = {}
    self.InputVariablesOrder = []
    self.OutputVariablesOrder = []
    self.observers_dict = {}

    self.debug = debug
    if self.debug:
      logging.getLogger().setLevel(logging.DEBUG)
    else:
      logging.getLogger().setLevel(logging.WARNING)

    # Observation Management
    self.ObservationOperatorType = {}
    self.FunctionObservationOperator = {}

    # Evolution Management
    self.EvolutionModelType = {}
    self.FunctionEvolutionModel = {}

  def setYIInputVariable(self, name, size):
    self.InputVariables[name] = size
    self.InputVariablesOrder.append(name)

  def setYIOutputVariable(self, name, size):
    self.OutputVariables[name] = size
    self.OutputVariablesOrder.append(name)

  def setYIAlgorithmParameters(self, parameters):
    self.algorithm_dict = parameters

  def initYIAlgorithm(self):
    self.ADD.setAlgorithmParameters(Algorithm=self.algorithm)
    if self.algorithm_dict != None:
      logging.debug("DASTUDY AlgorithmParameters: "+str(self.algorithm_dict))
      self.ADD.updateAlgorithmParameters(Parameters=self.algorithm_dict)

  def YI_prepare_to_pickle(self):
    return self.ADD.prepare_to_pickle()

  #--------------------------------------

  def __dir__(self):
    return ['getResults', '__doc__', '__init__', '__module__']

  def getResults(self):
    "Unique méthode à ne pas inclure YI"
    return self.ADD

  #--------------------------------------
  # Methods to initialize assimilation study

  def setYIBackgroundType(self, Type):
    if Type == "Vector":
      self.BackgroundType = Type
    else:
      raise daError("[daStudy::setYIBackgroundType] The following type is unkown : %s. Authorized types are : Vector"%(Type,))

  def setYIBackgroundStored(self, Stored):
    if Stored:
      self.BackgroundStored = True
    else:
      self.BackgroundStored = False

  def setYIBackground(self, Background):
    try:
      self.BackgroundType
      self.BackgroundStored
    except AttributeError:
      raise daError("[daStudy::setYIBackground] Type or Storage is not defined !")
    self.Background = Background
    if self.BackgroundType == "Vector":
      self.ADD.setBackground(Vector = Background, Stored = self.BackgroundStored)

  def getYIBackground(self):
    return self.Background

  #--------------------------------------

  def setYICheckingPointType(self, Type):
    if Type == "Vector":
      self.CheckingPointType = Type
    else:
      raise daError("[daStudy::setYICheckingPointType] The following type is unkown : %s. Authorized types are : Vector"%(Type,))

  def setYICheckingPointStored(self, Stored):
    if Stored:
      self.CheckingPointStored = True
    else:
      self.CheckingPointStored = False

  def setYICheckingPoint(self, CheckingPoint):
    try:
      self.CheckingPointType
      self.CheckingPointStored
    except AttributeError:
      raise daError("[daStudy::setYICheckingPoint] Type or Storage is not defined !")
    self.CheckingPoint = CheckingPoint
    if self.CheckingPointType == "Vector":
      self.ADD.setBackground(Vector = CheckingPoint, Stored = self.CheckingPointStored)

  #--------------------------------------

  def setYIBackgroundErrorType(self, Type):
    if Type in ("Matrix", "ScalarSparseMatrix", "DiagonalSparseMatrix"):
      self.BackgroundErrorType = Type
    else:
      raise daError("[daStudy::setYIBackgroundErrorType] The following type is unkown : %s. Authorized types are : Matrix, ScalarSparseMatrix, DiagonalSparseMatrix"%(Type,))

  def setYIBackgroundErrorStored(self, Stored):
    if Stored:
      self.BackgroundErrorStored = True
    else:
      self.BackgroundErrorStored = False

  def setYIBackgroundError(self, BackgroundError):
    try:
      self.BackgroundErrorType
      self.BackgroundErrorStored
    except AttributeError:
      raise daError("[daStudy::setYIBackgroundError] Type or Storage is not defined !")
    if self.BackgroundErrorType == "Matrix":
      self.ADD.setBackgroundError(Matrix = BackgroundError, Stored = self.BackgroundErrorStored)
    if self.BackgroundErrorType == "ScalarSparseMatrix":
      self.ADD.setBackgroundError(ScalarSparseMatrix = BackgroundError, Stored = self.BackgroundErrorStored)
    if self.BackgroundErrorType == "DiagonalSparseMatrix":
      self.ADD.setBackgroundError(DiagonalSparseMatrix = BackgroundError, Stored = self.BackgroundErrorStored)

  #--------------------------------------

  def setYIControlInputType(self, Type):
    if Type in ("Vector", "VectorSerie"):
      self.ControlInputType = Type
    else:
      raise daError("[daStudy::setYIControlInputType] The following type is unkown : %s. Authorized types are : Vector, VectorSerie"%(Type,))

  def setYIControlInputStored(self, Stored):
    if Stored:
      self.ControlInputStored = True
    else:
      self.ControlInputStored = False

  def setYIControlInput(self, ControlInput):
    try:
      self.ControlInputType
      self.ControlInputStored
    except AttributeError:
      raise daError("[daStudy::setYIControlInput] Type or Storage is not defined !")
    if self.ControlInputType == "Vector":
      self.ADD.setControlInput(Vector = ControlInput, Stored = self.ControlInputStored)
    if self.ControlInputType == "VectorSerie":
      self.ADD.setControlInput(VectorSerie = ControlInput, Stored = self.ControlInputStored)

  #--------------------------------------

  def setYIObservationType(self, Type):
    if Type in ("Vector", "VectorSerie"):
      self.ObservationType = Type
    else:
      raise daError("[daStudy::setYIObservationType] The following type is unkown : %s. Authorized types are : Vector, VectorSerie"%(Type,))

  def setYIObservationStored(self, Stored):
    if Stored:
      self.ObservationStored = True
    else:
      self.ObservationStored = False

  def setYIObservation(self, Observation):
    try:
      self.ObservationType
      self.ObservationStored
    except AttributeError:
      raise daError("[daStudy::setYIObservation] Type or Storage is not defined !")
    if self.ObservationType == "Vector":
      self.ADD.setObservation(Vector = Observation, Stored = self.ObservationStored)
    if self.ObservationType == "VectorSerie":
      self.ADD.setObservation(VectorSerie = Observation, Stored = self.ObservationStored)

  #--------------------------------------

  def setYIObservationErrorType(self, Type):
    if Type in ("Matrix", "ScalarSparseMatrix", "DiagonalSparseMatrix"):
      self.ObservationErrorType = Type
    else:
      raise daError("[daStudy::setYIObservationErrorType] The following type is unkown : %s. Authorized types are : Matrix, ScalarSparseMatrix, DiagonalSparseMatrix"%(Type,))

  def setYIObservationErrorStored(self, Stored):
    if Stored:
      self.ObservationErrorStored = True
    else:
      self.ObservationErrorStored = False

  def setYIObservationError(self, ObservationError):
    try:
      self.ObservationErrorType
      self.ObservationErrorStored
    except AttributeError:
      raise daError("[daStudy::setYIObservationError] Type or Storage is not defined !")
    if self.ObservationErrorType == "Matrix":
      self.ADD.setObservationError(Matrix = ObservationError, Stored = self.ObservationErrorStored)
    if self.ObservationErrorType == "ScalarSparseMatrix":
      self.ADD.setObservationError(ScalarSparseMatrix = ObservationError, Stored = self.ObservationErrorStored)
    if self.ObservationErrorType == "DiagonalSparseMatrix":
      self.ADD.setObservationError(DiagonalSparseMatrix = ObservationError, Stored = self.ObservationErrorStored)

  #--------------------------------------

  def getYIObservationOperatorType(self, Name):
    rtn = None
    try:
      rtn = self.ObservationOperatorType[Name]
    except:
      pass
    return rtn

  def setYIObservationOperatorType(self, Name, Type):
    if Type in ("Matrix", "Function"):
      self.ObservationOperatorType[Name] = Type
    else:
      raise daError("[daStudy::setYIObservationOperatorType] The following type is unkown : %s. Authorized types are : Matrix, Function"%(Type,))

  def setYIObservationOperator(self, Name, ObservationOperator):
    try:
      self.ObservationOperatorType[Name]
    except AttributeError:
      raise daError("[daStudy::setYIObservationOperator] Type is not defined !")
    if self.ObservationOperatorType[Name] == "Matrix":
      self.ADD.setObservationOperator(Matrix = ObservationOperator)
    elif self.ObservationOperatorType[Name] == "Function":
      self.FunctionObservationOperator[Name] = ObservationOperator

  #--------------------------------------

  def setYIEvolutionErrorType(self, Type):
    if Type in ("Matrix", "ScalarSparseMatrix", "DiagonalSparseMatrix"):
      self.EvolutionErrorType = Type
    else:
      raise daError("[daStudy::setYIEvolutionErrorType] The following type is unkown : %s. Authorized types are : Matrix, ScalarSparseMatrix, DiagonalSparseMatrix"%(Type,))

  def setYIEvolutionErrorStored(self, Stored):
    if Stored:
      self.EvolutionErrorStored = True
    else:
      self.EvolutionErrorStored = False

  def setYIEvolutionError(self, EvolutionError):
    try:
      self.EvolutionErrorType
      self.EvolutionErrorStored
    except AttributeError:
      raise daError("[daStudy::setYIEvolutionError] Type or Storage is not defined !")
    if self.EvolutionErrorType == "Matrix":
      self.ADD.setEvolutionError(Matrix = EvolutionError, Stored = self.EvolutionErrorStored)
    if self.EvolutionErrorType == "ScalarSparseMatrix":
      self.ADD.setEvolutionError(ScalarSparseMatrix = EvolutionError, Stored = self.EvolutionErrorStored)
    if self.EvolutionErrorType == "DiagonalSparseMatrix":
      self.ADD.setEvolutionError(DiagonalSparseMatrix = EvolutionError, Stored = self.EvolutionErrorStored)

  #--------------------------------------

  def getYIEvolutionModelType(self, Name):
    rtn = None
    try:
      rtn = self.EvolutionModelType[Name]
    except:
      pass
    return rtn

  def setYIEvolutionModelType(self, Name, Type):
    if Type in ("Matrix", "Function"):
      self.EvolutionModelType[Name] = Type
    else:
      raise daError("[daStudy::setYIEvolutionModelType] The following type is unkown : %s. Authorized types are : Matrix, Function"%(Type,))

  def setYIEvolutionModel(self, Name, EvolutionModel):
    try:
      self.EvolutionModelType[Name]
    except AttributeError:
      raise daError("[daStudy::setYIEvolutionModel] Type is not defined !")
    if self.EvolutionModelType[Name] == "Matrix":
      self.ADD.setEvolutionModel(Matrix = EvolutionModel)
    elif self.EvolutionModelType[Name] == "Function":
      self.FunctionEvolutionModel[Name] = EvolutionModel

  #--------------------------------------

  def addYIObserver(self, name, scheduler, info, number):
    self.observers_dict[name] = {}
    self.observers_dict[name]["scheduler"] = scheduler
    self.observers_dict[name]["info"] = info
    self.observers_dict[name]["number"] = number




# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# Copyright (C) 2007-2008  CEA/DEN, EDF R&D, OPEN CASCADE
#
# Copyright (C) 2003-2007  OPEN CASCADE, EADS/CCR, LIP6, CEA/DEN,
# CEDRAT, EDF R&D, LEG, PRINCIPIA R&D, BUREAU VERITAS
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#

__all__ = [
    "moduleID",
    "objectID",
    "unknownID",
    "componentName",
    "modulePixmap",
    "verbose",
    "getORB",
    "getLCC",
    "getStudyManager",
    "getEngine",
    "getEngineIOR",
    "findOrCreateComponent",
    "getObjectID",
    ]

import os
from omniORB import CORBA
from SALOME_NamingServicePy import SALOME_NamingServicePy_i
from LifeCycleCORBA import LifeCycleCORBA
import salome
import SALOMEDS
import SALOMEDS_Attributes_idl
from salome.kernel.studyedit import getStudyEditor

#import OMA_ORB

###
# Get OMA module's ID
###
def moduleID():
    MODULE_ID = 6100
    return MODULE_ID

###
# Get OMA object's ID
###
def objectID():
    OBJECT_ID = 6110
    return OBJECT_ID

###
# Get unknown ID
###
def unknownID():
    FOREIGN_ID = -1
    return FOREIGN_ID

def componentName():
    """
    This provide the name of the module component to be associated to the study.
    """
    # Note that this name should be (i) the name used for the class implementing
    # the component CORBA interface and (ii) the name used to declare the component
    # in the catalog of the module.
    return "ADAO"

# _MEM_ we use here the tr() translation methode to manage constant parameters
# in the application. We could have specified instead constant values directly
# in the code below. It's a matter of convenience.
from PyQt5.QtCore import QObject
QObjectTR=QObject()

def componentUserName():
    return "ADAO"

def modulePixmap():
    """
    Get the reference pixmap for this module.
    """
    return "ADAO_small.png"

def studyItemPixmapOk():
    """
    Get the reference pixmap for items of this module.
    """
    return "ADAO_small_vert.png"

def studyItemPixmapNOk():
    """
    Get the reference pixmap for items of this module.
    """
    return "ADAO_small_rouge.png"

__verbose__ = None
def verbose():
    global __verbose__
    if __verbose__ is None:
        try:
            __verbose__ = int( os.getenv( 'SALOME_VERBOSE', 0 ) )
        except:
            __verbose__ = 0
            pass
        pass
    return __verbose__

###
# Get ORB reference
###
def getORB():
    salome.salome_init()
    return salome.orb

##
# Get life cycle CORBA instance
##
def getLCC():
    salome.salome_init()
    return salome.lcc

##
# Get study manager
###
def getStudyManager():
    salome.salome_init()
    return salome.myStudy

###
# Get OMA engine
###
__engine__ = None
def getEngine():
    global __engine__
    if __engine__ is None:
        __engine__ = getLCC().FindOrLoadComponent( "FactoryServer", componentName() )
        pass
    return __engine__

###
# Get OMA engine IOR
###
def getEngineIOR():
    IOR = ""
    if getORB() and getEngine():
        IOR = getORB().object_to_string( getEngine() )
        pass
    return IOR

###
# Find or create OMA component object in a study
###
def findOrCreateComponent( study ):
    father = study.FindComponent( componentName() )
    if father is None:
        builder = study.NewBuilder()
        father = builder.NewComponent( componentName() )
        attr = builder.FindOrCreateAttribute( father, "AttributeName" )
        attr.SetValue( componentName() )
        attr = builder.FindOrCreateAttribute( father, "AttributePixMap" )
        attr.SetPixMap( modulePixmap() )
        attr = builder.FindOrCreateAttribute( father, "AttributeLocalID" )
        attr.SetValue( moduleID() )
        try:
            builder.DefineComponentInstance( father, getEngine() )
            pass
        except:
            pass
        pass
    return father

###
# Get object's ID
###
def getObjectID( entry ): # study, entry ):
    ID = unknownID()
    if entry: # study and entry:
        sobj = getStudyEditor().study.FindObjectID( entry )
        if sobj is not None:
            test, anAttr = sobj.FindAttribute( "AttributeLocalID" )
            if test: ID = anAttr._narrow( SALOMEDS.AttributeLocalID ).Value()
            pass
        pass
    return




# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import numpy, logging, copy
from daCore import BasicObjects

# ==============================================================================
class ElementaryAlgorithm(BasicObjects.Algorithm):
    def __init__(self):
        BasicObjects.Algorithm.__init__(self, "PARTICLESWARMOPTIMIZATION")
        self.defineRequiredParameter(
            name     = "MaximumNumberOfIterations",
            default  = 50,
            typecast = int,
            message  = "Nombre maximal de pas d'optimisation",
            minval   = 0,
            oldname  = "MaximumNumberOfSteps",
            )
        self.defineRequiredParameter(
            name     = "MaximumNumberOfFunctionEvaluations",
            default  = 15000,
            typecast = int,
            message  = "Nombre maximal d'évaluations de la fonction",
            minval   = -1,
            )
        self.defineRequiredParameter(
            name     = "SetSeed",
            typecast = numpy.random.seed,
            message  = "Graine fixée pour le générateur aléatoire",
            )
        self.defineRequiredParameter(
            name     = "NumberOfInsects",
            default  = 100,
            typecast = int,
            message  = "Nombre d'insectes dans l'essaim",
            minval   = -1,
            )
        self.defineRequiredParameter(
            name     = "SwarmVelocity",
            default  = 1.,
            typecast = float,
            message  = "Vitesse de groupe imposée par l'essaim",
            minval   = 0.,
            )
        self.defineRequiredParameter(
            name     = "GroupRecallRate",
            default  = 0.5,
            typecast = float,
            message  = "Taux de rappel au meilleur insecte du groupe (entre 0 et 1)",
            minval   = 0.,
            maxval   = 1.,
            )
        self.defineRequiredParameter(
            name     = "QualityCriterion",
            default  = "AugmentedWeightedLeastSquares",
            typecast = str,
            message  = "Critère de qualité utilisé",
            listval  = [
                "AugmentedWeightedLeastSquares", "AWLS", "DA",
                "WeightedLeastSquares", "WLS",
                "LeastSquares", "LS", "L2",
                "AbsoluteValue", "L1",
                "MaximumError", "ME",
                ],
            )
        self.defineRequiredParameter(
            name     = "StoreInternalVariables",
            default  = False,
            typecast = bool,
            message  = "Stockage des variables internes ou intermédiaires du calcul",
            )
        self.defineRequiredParameter(
            name     = "StoreSupplementaryCalculations",
            default  = [],
            typecast = tuple,
            message  = "Liste de calculs supplémentaires à stocker et/ou effectuer",
            listval  = [
                "Analysis",
                "BMA",
                "CostFunctionJ",
                "CostFunctionJb",
                "CostFunctionJo",
                "CurrentIterationNumber",
                "CurrentState",
                "Innovation",
                "OMA",
                "OMB",
                "SimulatedObservationAtBackground",
                "SimulatedObservationAtCurrentState",
                "SimulatedObservationAtOptimum",
                ]
            )
        self.defineRequiredParameter( # Pas de type
            name     = "BoxBounds",
            message  = "Liste des valeurs de bornes d'incréments de paramètres",
            )
        self.requireInputArguments(
            mandatory= ("Xb", "Y", "HO", "R", "B"),
            )
        self.setAttributes(tags=(
            "Optimization",
            "NonLinear",
            "MetaHeuristic",
            "Population",
            ))

    def run(self, Xb=None, Y=None, U=None, HO=None, EM=None, CM=None, R=None, B=None, Q=None, Parameters=None):
        self._pre_run(Parameters, Xb, Y, U, HO, EM, CM, R, B, Q)
        #
        if ("BoxBounds" in self._parameters) and isinstance(self._parameters["BoxBounds"], (list, tuple)) and (len(self._parameters["BoxBounds"]) > 0):
            BoxBounds = self._parameters["BoxBounds"]
            logging.debug("%s Prise en compte des bornes d'incréments de paramètres effectuée"%(self._name,))
        else:
            raise ValueError("Particle Swarm Optimization requires bounds on all variables increments to be truly given (BoxBounds).")
        BoxBounds   = numpy.array(BoxBounds)
        if numpy.isnan(BoxBounds).any():
            raise ValueError("Particle Swarm Optimization requires bounds on all variables increments to be truly given (BoxBounds), \"None\" is not allowed. The actual increments bounds are:\n%s"%BoxBounds)
        #
        Phig = float( self._parameters["GroupRecallRate"] )
        Phip = 1. - Phig
        logging.debug("%s Taux de rappel au meilleur insecte du groupe (entre 0 et 1) = %s et à la meilleure position précédente (son complémentaire à 1) = %s"%(self._name, str(Phig), str(Phip)))
        #
        Hm = HO["Direct"].appliedTo
        #
        BI = B.getI()
        RI = R.getI()
        #
        def CostFunction(x, QualityMeasure="AugmentedWeightedLeastSquares"):
            _X  = numpy.ravel( x ).reshape((-1,1))
            _HX = numpy.ravel( Hm( _X ) ).reshape((-1,1))
            _Innovation = Y - _HX
            #
            if QualityMeasure in ["AugmentedWeightedLeastSquares","AWLS","DA"]:
                if BI is None or RI is None:
                    raise ValueError("Background and Observation error covariance matrices has to be properly defined!")
                Jb  = 0.5 * (_X - Xb).T @ (BI @ (_X - Xb))
                Jo  = 0.5 * _Innovation.T @ (RI @ _Innovation)
            elif QualityMeasure in ["WeightedLeastSquares","WLS"]:
                if RI is None:
                    raise ValueError("Observation error covariance matrix has to be properly defined!")
                Jb  = 0.
                Jo  = 0.5 * _Innovation.T @ (RI @ _Innovation)
            elif QualityMeasure in ["LeastSquares","LS","L2"]:
                Jb  = 0.
                Jo  = 0.5 * _Innovation.T @ _Innovation
            elif QualityMeasure in ["AbsoluteValue","L1"]:
                Jb  = 0.
                Jo  = numpy.sum( numpy.abs(_Innovation) )
            elif QualityMeasure in ["MaximumError","ME"]:
                Jb  = 0.
                Jo  = numpy.max( numpy.abs(_Innovation) )
            #
            J   = float( Jb ) + float( Jo )
            #
            return J
        #
        if Xb is not None:
            Xini = numpy.ravel(Xb)
        else:
            Xini = numpy.zeros(len(BoxBounds[:,0]))
        #
        SpaceUp  = BoxBounds[:,1] + Xini
        SpaceLow = BoxBounds[:,0] + Xini
        nbparam  = len(SpaceUp)
        #
        # Initialisation de l'essaim
        # --------------------------
        NumberOfFunctionEvaluations = 0
        LimitVelocity = numpy.abs(SpaceUp-SpaceLow)
        #
        PosInsect = []
        VelocityInsect = []
        for i in range(nbparam) :
            PosInsect.append(numpy.random.uniform(low=SpaceLow[i], high=SpaceUp[i], size=self._parameters["NumberOfInsects"]))
            VelocityInsect.append(numpy.random.uniform(low=-LimitVelocity[i], high=LimitVelocity[i], size=self._parameters["NumberOfInsects"]))
        VelocityInsect = numpy.array(VelocityInsect)
        PosInsect = numpy.array(PosInsect)
        #
        BestPosInsect = numpy.array(PosInsect)
        qBestPosInsect = []
        _Best = copy.copy(SpaceLow)
        _qualityBest = CostFunction(_Best,self._parameters["QualityCriterion"])
        NumberOfFunctionEvaluations += 1
        #
        for i in range(self._parameters["NumberOfInsects"]):
            insect  = numpy.ravel(PosInsect[:,i])
            quality = CostFunction(insect,self._parameters["QualityCriterion"])
            NumberOfFunctionEvaluations += 1
            qBestPosInsect.append(quality)
            if quality < _qualityBest:
                _Best  = copy.copy( insect )
                _qualityBest = copy.copy( quality )
        logging.debug("%s Initialisation, Insecte = %s, Qualité = %s"%(self._name, str(_Best), str(_qualityBest)))
        #
        self.StoredVariables["CurrentIterationNumber"].store( len(self.StoredVariables["CostFunctionJ"]) )
        if self._parameters["StoreInternalVariables"] or self._toStore("CurrentState"):
            self.StoredVariables["CurrentState"].store( _Best )
        self.StoredVariables["CostFunctionJb"].store( 0. )
        self.StoredVariables["CostFunctionJo"].store( 0. )
        self.StoredVariables["CostFunctionJ" ].store( _qualityBest )
        #
        # Minimisation de la fonctionnelle
        # --------------------------------
        for n in range(self._parameters["MaximumNumberOfIterations"]):
            for i in range(self._parameters["NumberOfInsects"]) :
                insect  = numpy.ravel(PosInsect[:,i])
                rp = numpy.random.uniform(size=nbparam)
                rg = numpy.random.uniform(size=nbparam)
                for j in range(nbparam) :
                    VelocityInsect[j,i] = self._parameters["SwarmVelocity"]*VelocityInsect[j,i] +  Phip*rp[j]*(BestPosInsect[j,i]-PosInsect[j,i]) +  Phig*rg[j]*(_Best[j]-PosInsect[j,i])
                    PosInsect[j,i] = PosInsect[j,i]+VelocityInsect[j,i]
                quality = CostFunction(insect,self._parameters["QualityCriterion"])
                NumberOfFunctionEvaluations += 1
                if quality < qBestPosInsect[i]:
                    BestPosInsect[:,i] = copy.copy( insect )
                    qBestPosInsect[i]  = copy.copy( quality )
                    if quality < _qualityBest :
                        _Best  = copy.copy( insect )
                        _qualityBest = copy.copy( quality )
            logging.debug("%s Etape %i, Insecte = %s, Qualité = %s"%(self._name, n, str(_Best), str(_qualityBest)))
            #
            self.StoredVariables["CurrentIterationNumber"].store( len(self.StoredVariables["CostFunctionJ"]) )
            if self._parameters["StoreInternalVariables"] or self._toStore("CurrentState"):
                self.StoredVariables["CurrentState"].store( _Best )
            if self._toStore("SimulatedObservationAtCurrentState"):
                _HmX = Hm( _Best )
                self.StoredVariables["SimulatedObservationAtCurrentState"].store( _HmX )
            self.StoredVariables["CostFunctionJb"].store( 0. )
            self.StoredVariables["CostFunctionJo"].store( 0. )
            self.StoredVariables["CostFunctionJ" ].store( _qualityBest )
            if NumberOfFunctionEvaluations > self._parameters["MaximumNumberOfFunctionEvaluations"]:
                logging.debug("%s Stopping search because the number %i of function evaluations is exceeding the maximum %i."%(self._name, NumberOfFunctionEvaluations, self._parameters["MaximumNumberOfFunctionEvaluations"]))
                break
        #
        # Obtention de l'analyse
        # ----------------------
        Xa = _Best
        #
        self.StoredVariables["Analysis"].store( Xa )
        #
        # Calculs et/ou stockages supplémentaires
        # ---------------------------------------
        if self._toStore("OMA") or \
            self._toStore("SimulatedObservationAtOptimum"):
            HXa = Hm(Xa)
        if self._toStore("Innovation") or \
            self._toStore("OMB") or \
            self._toStore("SimulatedObservationAtBackground"):
            HXb = Hm(Xb)
            Innovation = Y - HXb
        if self._toStore("Innovation"):
            self.StoredVariables["Innovation"].store( Innovation )
        if self._toStore("OMB"):
            self.StoredVariables["OMB"].store( Innovation )
        if self._toStore("BMA"):
            self.StoredVariables["BMA"].store( numpy.ravel(Xb) - numpy.ravel(Xa) )
        if self._toStore("OMA"):
            self.StoredVariables["OMA"].store( numpy.ravel(Y) - numpy.ravel(HXa) )
        if self._toStore("SimulatedObservationAtBackground"):
            self.StoredVariables["SimulatedObservationAtBackground"].store( HXb )
        if self._toStore("SimulatedObservationAtOptimum"):
            self.StoredVariables["SimulatedObservationAtOptimum"].store( HXa )
        #
        self._post_run(HO)
        return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')
